/**************************************************************************
 *
 * $Id: crypt.h,v 1.6 2005/11/29 20:15:46 pguio Exp $
 *
 * Copyright (c) Members of the EGEE Collaboration. 2004.
 * See http://eu-egee.org/partners/ for details on the copyright
 * holders.
 * For license conditions see the license file or
 * http://eu-egee.org/license.html
 * 
 *
 * Authors: Patrick Guio <patrick.guio@bccs.uib.no>
 *
 **************************************************************************/

/** \mainpage Cryptographic Tools for glite EDS
 *
 * \section Introduction
 *
 * This is the documentation of the security module that contain cryptographic tools
 * to be used in the gLite grid middleware.
 * 
 * \section License Conditions
 *
 * Copyright (c) Members of the EGEE Collaboration. 2004.
 *
 * See http://eu-egee.org/partners/ for details on the copyright
 * holders.
 *
 * For license conditions see the license file or
 * http://eu-egee.org/license.html
 *
 * \author Patrick Guio <patrick.guio@bccs.uib.no>
 */

/** \file crypt.h
 * Declarations for gLite encryted data storage.
 *
 * Provides declarations for a set of functionalities to process
 * files
 *   - buffer and file encryption/decryption,
 *   - key generation
 *   - key manipulation
 *      - split/join
 *      - base64-encode/decode
 *   - key i/o manipulation.
 *
 * This file is the only file to include.
 * 
 */


#ifndef CRYPT_H
#define CRYPT_H 1

#include <string>
#include <iostream>
#include <fstream>
#include <sstream>
#include <iomanip>
#include <stdexcept>
#include <vector>
#include <limits>

#include <openssl/bio.h>
#include <openssl/evp.h>
#include <openssl/err.h>
#include <openssl/rand.h>

namespace glite
{

  //! Class for encryted data storage
  /*!
   * Available methods to 
   * - encrypt/decrypt a buffer or a file using an encryption key
   *   and an algorithm implementation from OpenSSL.
   * - Base64 encode/decode a key.
   * - split/join key using Trivial Sharing Secret scheme.
   * - split/join key using Shamir Sharing Secret scheme.
   */
  class Crypt
  {

  public:

    //! default key type
    typedef unsigned char byte;
    //! container for key
    typedef std::vector<byte> Key;

    //! container for Base64 key.
    typedef std::string Base64Key;

    //! container for TSS split keys.
    typedef std::vector<Key> SplitKeys;

    //! key type for SSS key
    typedef unsigned short shortByte;
    //! container for SSS key
    typedef std::vector<shortByte> ShortKey;
    //! container for SSS polynom representation.
    typedef std::vector<shortByte> Polynom;
    //! container for SSS split keys.
    typedef std::vector<ShortKey > SplitShortKeys;
    //! key type for SSS key processing
    typedef unsigned long longByte;
    //! container for SSS key processing
    typedef std::vector<longByte> LongKey;

    //! default constructor
    Crypt(const std::string cipher="bf-cbc", int keyLength=0);
    //! default destructor
    ~Crypt();

    //! encrypts a data buffer
    void encrypt(byte *in, int isize, byte * &out, int &osize)                 const;
    //! encrypts a file
    void encrypt(const std::string &ifile, const std::string &ofile)           const;

    //! decrypts an input data buffer
    void decrypt(byte *in, int isize, byte * &out, int &osize)                 const;
    //! decrypts a file
    void decrypt(const std::string &ifile, const std::string &ofile)           const;

    //! Base64 encode an input key
    void encodeBase64(const Key & key, Base64Key & b64key)                     const;
    //! Base64 decode an input key
    void decodeBase64(const Base64Key & b64key, Key & key)                     const;

    //! Split Key using Trivial Secret Sharing
    void splitKeyTSS(const Key & key, int nShares, SplitKeys & keys)           const;
    //! Join Keys using Trivial Secret Sharing
    void joinKeyTSS(const SplitKeys & keys, Key & key)                         const;

    //! Split Key using Shamir Secret Sharing
    void splitKeySSS(const Key & key, int nNeeded, int nShares,
                     Key & x, SplitShortKeys & keys)                           const;
    //! Join Keys using Shamir Secret Sharing
    void joinKeySSS(const int nNeeded,
                    const Key & x, const SplitShortKeys & keys, Key & key)     const;

    //! Set an encryption/decryption key and initialisation vector (optional)
    void setKeyAndIV(const Key &key, const Key &iv=Key());
    //! Return current encryption/decryption key
    Key  getKey()                                                              const;
    //! Return current initialisation vector
    Key  getIV ()                                                              const;

    //! Set buffer size to encrypt/decrypt file
    void setBufferSize(int size);
    //! Return current buffer size
    int  getBufferSize() const;

    //! Set verbose level
    void setVerbose(int verbose);
    //! Return current verbose level
    int  getVerbose() const;

    //! Print state of Crypter object
    void printOn(std::ostream &os)                                            const;

  private:

    //! Seed the pseudo random number generator
    void seedPRNG(int bytes=1)                                                const;

    //! Draw a pseudo random key and set the encryption/decryption key
    void drawKey();
    //! Draw a pseudo-random initialisation vector and set the initialisation vector
    void drawIV();
    //! Initialise Crypter object for encryption/decryption
    void initCrypters();

    //! Handle error and throw an exception
    void handleError(const char *thisFile, int thisLine)                      const;

    //! Draw a random key (TSS)
    void drawPRN(Key & key)                                                   const;
    //! Draw a random polynom (SSS)
    void drawPRN(Polynom & p)                                                 const;

    //! Initialise a polynomial to be used in SSS
    Polynom  initPolynom(int k, byte a0)                                      const;
    //! Evaluate a polynomial at a given value (SSS)
    longByte evalPolynom(const Polynom & p, const longByte x)                 const;
    //! Calculate the modular (prime) inverse of a number
    longByte inverseModulo(const longByte n)                                  const;
    //! Compute the modular (prime) division of two numbers
    longByte divideModulo(const longByte numerator,
                          const longByte denominator)                         const;
    //! Compute Lagrange interpolation at zero
    void     evalLagrangeInterpAt0(const Key & x, LongKey & numerator,
                                   LongKey & denominator)                     const;


    const std::string m_cipherName; //!< name of OpenSSL cypher type
    unsigned int m_keyLength;       //!< encryption/decryption key length
    unsigned int m_ivLength;        //!< initialisation vector length

#if 0
    // Do not use Engines since support is currently
    // OpenSSL <  0.9.6 not defined
    // OpenSSL == 0.9.6 definef struct engine
    // OpenSSL >  0.9.6 defines struct ENGINE
    ENGINE * m_impl;                //!< cipher implementation
#endif

    int m_bufferSize;               //!< buffer size to encrypt/decrypt file

    EVP_CIPHER_CTX * m_ectx;        //!< OpenSSL encrypting cipher context
    EVP_CIPHER_CTX * m_dctx;        //!< OpenSSL decrypting cipher context

    const EVP_CIPHER * m_type;      //!< OpenSSL EVP_CIPHER structure

    Key m_key;                      //!< encryption/decryption key
    Key m_iv;                       //!< initialisation vector

    int m_verbose;                  //!< verbose level

  };

  //! Output insertion operator for object of class Crypt
  std::ostream & operator<<(std::ostream &os, const Crypt & c);

  //! Convert an array into a key
  template<typename key_type>
  key_type Array2Key(const typename key_type::value_type * array, unsigned int length);
  //! Convert a key into an array
  template<typename key_type>
  typename key_type::value_type * Key2Array(const key_type & vec , unsigned int &length);


  //! State structure manipulator to output key
  template<typename key_type>
  struct outputKey
  {
    outputKey(const key_type & key, const char separator=0, int lineWidth=64);
    ~outputKey();
    key_type key_;
    char separator_;
    int lineWidth_;
  };

  //! Output insertion operator for object of type key
  template<typename key_type>
  std::ostream & operator<<(std::ostream &os, const outputKey<std::vector<key_type> > &k);
  //! Output insertion operator for object of type Base64 key
  std::ostream & operator<<(std::ostream &os, const outputKey<Crypt::Base64Key> &k);
  //! Output insertion operator for object of type split keys
  template<typename key_type>
  std::ostream & operator<<(std::ostream &os, const outputKey<std::vector<std::vector<key_type> > > &k);

  //! State structure manipulator to input key
  template<typename key_type>
  struct inputKey
  {
    inputKey(const char separator=0);
    ~inputKey();
    key_type key_;
    char separator_;
  };

  //! Input insertion operator for object of type key
  template<typename key_type>
  std::istream & operator>>(std::istream & is, inputKey<key_type> &k);


  //! State structure manipulator to format key
  struct setFormat
  {
    setFormat(std::ios::fmtflags base, int width=0, char padding=' ');
    ~setFormat();
    std::ios::fmtflags base_;
    int width_;
    char padding_;
  };

  //! Input insertion operator for object of type format
  std::istream & operator>>(std::istream& is, const setFormat &f);
  //! Output insertion operator for object of type format
  std::ostream & operator<<(std::ostream& os, const setFormat &f);

}

#endif // CRYPT_H
